@@ -3,7 +3,9 @@ import math
 import torch
 import faiss
 import numpy as np
+
 from einops import rearrange
+from memorizing_transformers_pytorch.utils import rearrange_with_dim_list
 
 # constants
 
@@ -28,7 +30,7 @@ def count_intersect(x, y):
 # function takes in list of ids, which ranges from newest to oldest
 # and returns the list of ids that should be removed
 
-def expire_strategy_remove_oldest(ids, *, max_num_entries, num_hits):
+def expire_strategy_remove_oldest(ids, *, max_num_entries, num_hits, ages):
     ids_to_remove = ids[max_num_entries:]
     return ids_to_remove
 
@@ -65,6 +67,7 @@ class KNN():
     def reset(self):
         self.ids = np.empty((0,), dtype = np.int32)
         self.hits = np.empty((0,), dtype = np.int32)
+        self.age_num_iterations = np.empty((0,), dtype = np.int32)
         return self.index.reset()
 
     def train(self, x):
@@ -76,13 +79,22 @@ class KNN():
 
         self.ids = np.concatenate((ids, self.ids))
         self.hits = np.concatenate((np.zeros_like(ids), self.hits))
+        self.age_num_iterations = np.concatenate((np.zeros_like(ids), self.age_num_iterations))
 
         if self.cap_num_entries and len(self.ids) > self.max_num_entries:
-            remove_ids = self.expire_memory_fn(self.ids, max_num_entries = self.max_num_entries, num_hits = self.hits)
+
+            remove_ids = self.expire_memory_fn(
+                self.ids,
+                max_num_entries = self.max_num_entries,
+                num_hits = self.hits,
+                ages = self.age_num_iterations
+            )
+
             keep_mask = count_intersect(self.ids, remove_ids) == 0
 
             self.ids = self.ids[keep_mask]
             self.hits = self.hits[keep_mask]
+            self.age_num_iterations = self.age_num_iterations[keep_mask]
 
             assert len(self.ids) <= self.max_num_entries
             self.remove(remove_ids)
@@ -98,7 +110,8 @@ class KNN():
         topk,
         nprobe = 8,
         return_distances = False,
-        increment_hits = False
+        increment_hits = False,
+        increment_age = True
     ):
         if not self.index.is_trained:
             return np.full((x.shape[0], topk), -1)
@@ -110,6 +123,9 @@ class KNN():
             hits = count_intersect(self.ids, rearrange(indices, '... -> (...)'))
             self.hits += hits
 
+        if increment_age:
+            self.age_num_iterations += 1
+
         if return_distances:
             return indices, distances
 
@@ -173,9 +189,12 @@ class KNNMemory():
         queries,
         topk,
         nprobe = 8,
-        increment_hits = False
+        increment_hits = False,
+        increment_age = True
     ):
-        check_shape(queries, 'b n d', d = self.dim, b = self.num_indices)
+        _, *prec_dims, _ = queries.shape
+        check_shape(queries, 'b ... d', d = self.dim, b = self.num_indices)
+        queries = rearrange(queries, 'b ... d -> b (...) d')
 
         device = queries.device
         queries = queries.detach().cpu().numpy()
@@ -184,7 +203,7 @@ class KNNMemory():
         all_key_values = []
 
         for ind, (query, knn) in enumerate(zip(queries, self.knns)):
-            indices = knn.search(query, topk, nprobe, increment_hits = increment_hits)
+            indices = knn.search(query, topk, nprobe, increment_hits = increment_hits, increment_age = increment_age)
             mask = indices !=  -1
             db_indices = np.where(mask, indices, 0)
 
@@ -197,6 +216,9 @@ class KNNMemory():
         all_key_values = torch.stack(all_key_values)
         all_key_values = all_key_values.masked_fill(~rearrange(all_masks, '... -> ... 1 1'), 0.)
 
+        all_key_values = rearrange_with_dim_list(all_key_values, 'b (...p) ... -> b ...p ...', p = prec_dims)
+        all_masks = rearrange_with_dim_list(all_masks, 'b (...p) ... -> b ...p ...', p = prec_dims)
+
         return all_key_values.to(device), all_masks.to(device)
 
     def __del__(self):
